package controllers

import (
	"crypto/hmac"
	"crypto/sha1"
	"encoding/hex"
	"io"
	"log"
	"net/http"
	"strconv"
	"strings"
	"time"

	"github.com/gin-gonic/gin"
	"github.com/hibiken/asynq"

	"code.locsi.com/locsi/api/models"
	"code.locsi.com/locsi/queue/tasks"
	web_sub_models "code.locsi.com/locsi/websub/models"
)

// GET /sub/callback/:id
// Processes a WebSub subscription callback request.
func GetSubCallback(c *gin.Context) {
	log.Print("in get sub")

	hubMode, hubModeExists := c.GetQuery("hub.mode")
	if hubModeExists && hubMode == "subscribe" {
		// Parse the request
		var input web_sub_models.Hub
		if err := c.ShouldBind(&input); err != nil {
			log.Printf("binding error: %v", err.Error())

			c.JSON(http.StatusBadRequest, gin.H{"error": err.Error()})
			return
		}

		var PodcastSeries models.PodcastSeries
		if err := models.DB.Where("id = ?", c.Param("id")).First(&PodcastSeries).Error; err != nil {
			// The ID was invalid
			c.JSON(http.StatusBadRequest, gin.H{"error": "Unknown subscription"})
			return
		} else {
			// Store the lease time in the database
			leaseSeconds, leaseSecondsExists := c.GetQuery("hub.lease_seconds")
			if leaseSecondsExists {
				leaseSecondsInt, err := strconv.Atoi(leaseSeconds)
				if err == nil {
					models.DB.Model(&PodcastSeries).Update("web_sub_lease_expiration", time.Now().Add(time.Second*time.Duration(int(leaseSecondsInt))))
				}
			}

			// Send our response
			c.String(http.StatusOK, input.Challenge)
		}
	} else {
		// This is an invalid request
		c.JSON(http.StatusBadRequest, gin.H{"error": "Unknown subscription"})
		return
	}
}

func PostSubCallback(c *gin.Context) {
	log.Print("in post sub")

	// See if this id is using WebSub
	var PodcastSeries models.PodcastSeries
	if err := models.DB.Where("id = ?", c.Param("id")).First(&PodcastSeries).Error; err != nil {
		// The ID was invalid
		c.JSON(http.StatusBadRequest, gin.H{"error": "Unknown subscription"})
		return
	} else {
		// Read and store the response body.  We limit it to 20000 bytes to prevent blowing up our task queue if someone sends a bunch of data to this endpoint.
		limitedReader := &io.LimitedReader{R: c.Request.Body, N: 20000}
		body, err := io.ReadAll(limitedReader)
		if err != nil {
			body = nil
		} else {
			// See if this request was signed correctly
			gotHash := strings.SplitN(c.Request.Header.Get("X-Hub-Signature"), "=", 2)
			// TODO: Consider extending to support more algorithims
			if gotHash[0] == "sha1" {
				// Process the body of this request
				hash := hmac.New(sha1.New, []byte(PodcastSeries.WebSubSecret))
				if _, err := hash.Write(body); err != nil {
					// Cannot compute the HMAC for the request
					body = nil
				}
				expectedHash := hex.EncodeToString(hash.Sum(nil))

				if gotHash[1] != expectedHash {
					body = nil
				}
			}
		}

		// Add a task to crawl this feed for new data
		const redisAddr = "localhost:6379"
		client := asynq.NewClient(asynq.RedisClientOpt{Addr: redisAddr})
		defer client.Close()
		task, err := tasks.NewUpdatePodcastFromFeedTask(PodcastSeries.ID, string(body))
		if err != nil {
			log.Fatalf("could not create task: %v", err)
		}
		info, err := client.Enqueue(task)
		if err != nil {
			log.Fatalf("could not enqueue task: %v", err)
		}
		log.Printf("enqueued task: id=%s queue=%s", info.ID, info.Queue)
	}
}
